program cross-reference;


{   Cross-Reference Generator

    Author:          Peter Grogono
    Date Written:    24 September 1977
    Last Modified:   30 September 1977

    From: Programming in Pascal
          by Peter Grogono

    Converted to Turbo Pascal 2.0b
    by David W. Carroll
       May 5, 1985

}

CONST
  maxlines = maxint;
  maxwordlen = 20;
  maxlinelen = 80;
  maxonpage = 60;
  headingsize = 3;
  entrygap = 1;
  numbergap = 2;
  radix = 10;
  blank = ' ';
  cha = 'a';
  chz = 'z';
  heading = 'cross-reference listing';

TYPE
  counter = 1 .. maxlines;
  wordindex = 1 .. maxwordlen;
  pageindex = 1 .. maxonpage;
  lineindex = 1 .. maxlinelen;
  wordtype = ARRAY [wordindex] OF char;

  queuepointer = ^queueitem;
  queueitem =
    RECORD
      linenumber : counter;
      nextinqueue : queuepointer
    END;

  entrytype =
    RECORD
      wordvalue : wordtype;
      firstinqueue, lastinqueue : queuepointer
    END;

  treepointer = ^node;
  node =
    RECORD
      entry : entrytype;
      left,right : treepointer
    END;

VAR
  wordtree : treepointer;
  letters : SET OF char;

PROCEDURE buildtree (VAR tree : treepointer);
  VAR
    currentword : wordtype;
    currentline : counter;

  PROCEDURE getword (VAR word : wordtype;
                     VAR line : counter);

    VAR
      currentchar : char;
      index, blankindex : 0 .. maxwordlen;

    PROCEDURE getchar (VAR ch : char;
                       VAR lin : counter);

      BEGIN {getchar}
        IF eof(infile) THEN
          ch := blank
        ELSE IF eoln(infile) THEN
          BEGIN
            ch := blank;
            lin := lin + 1;
            readln(infile)
          END
        ELSE
          read(infile,ch)
      END; {getchar}

    BEGIN {getword}
      currentchar := blank;
      WHILE NOT (eof(infile) OR (currentchar IN letters)) DO
        getchar(currentchar,line);
      IF NOT eof(infile) THEN
        BEGIN
          index := 0;
          WHILE currentchar IN letters DO
            BEGIN
              IF index < maxwordlen THEN
                BEGIN
                  index := index + 1;
                  word[index] := currentchar
                END;
              IF eof(infile) THEN
                currentchar := blank
              ELSE getchar(currentchar,line)
            END; {while}
          IF index < maxwordlen THEN
            FOR blankindex := index + 1 TO maxwordlen DO
              word[blankindex] := blank
        END
    END; {getword}

  PROCEDURE entertree (VAR subtree : treepointer;
                       word : wordtype;
                       line : counter);

    VAR
      nextitem : queuepointer;

    BEGIN {entertree}
      IF subtree = NIL THEN
        BEGIN  {create a new entry}
          new(subtree);
          WITH subtree^ DO
            BEGIN
              left := NIL;
              right := NIL;
              WITH entry DO
                BEGIN
                  wordvalue := word;
                  new(firstinqueue);
                  lastinqueue := firstinqueue;
                  WITH firstinqueue^ DO
                    BEGIN
                      linenumber := line;
                      nextinqueue := NIL
                    END {with firstinqueue^}
                END {with entry}
            END {with subtree^}
        END
      ELSE {append a list item}
        WITH subtree^, entry DO
          IF word = wordvalue THEN
            BEGIN
              IF lastinqueue^.linenumber <> line THEN
                BEGIN
                  new(nextitem);
                  WITH nextitem^ DO
                    BEGIN
                      linenumber := line;
                      nextinqueue := NIL
                    END; {with nextitem^}
                  lastinqueue^.nextinqueue := nextitem;
                  lastinqueue := nextitem
                END
            END
          ELSE IF word < wordvalue THEN
            entertree(left,word,line)
          ELSE
            entertree(right,word,line)
    END; {entertree}

  BEGIN {buildtree}
    currentline := 1;
    WHILE NOT eof(infile) DO
      BEGIN
        getword(currentword,currentline);
        IF NOT eof(infile) THEN
          entertree(tree,currentword,currentline)
      END {while}
  END; {buildtree}

PROCEDURE printtree (tree : treepointer);
  VAR
    pageposition : pageindex;
    numberwidth, maxonline : lineindex;

  PROCEDURE printentry (subtree : treepointer;
                        VAR position : pageindex);

    VAR
      index : wordindex;
      itemcount : 0 .. maxlinelen;
      itemptr : queuepointer;

    PROCEDURE printline (VAR currentposition : pageindex;
                         newlines : pageindex);

      VAR
        linecounter : pageindex;

      BEGIN {printline}
        IF currentposition + newlines < maxonpage THEN
          BEGIN
            FOR linecounter := 1 TO newlines DO
              writeln;
            currentposition := currentposition + newlines
          END
        ELSE
          BEGIN
            writeln(chr(12));
            writeln(heading);
            FOR linecounter := 1 to headingsize -1 DO
              writeln;
            currentposition := headingsize + 1
          END
      END; {printline}

    BEGIN {printentry}
      IF subtree <> NIL THEN
        WITH subtree^ DO
          BEGIN
            printentry(left,position);
            printline(position,entrygap + 1);
            WITH entry DO
              BEGIN
                FOR index := 1 TO maxwordlen DO
                  write(wordvalue[index]);
                itemcount := 0;
                itemptr := firstinqueue;
                WHILE itemptr <> NIL DO
                  BEGIN
                    itemcount := itemcount + 1;
                    IF itemcount > maxonline THEN
                      BEGIN
                        printline(position,1);
                        write(blank : maxwordlen);
                        itemcount := 1
                      END;
                    write(itemptr^.linenumber : numberwidth);
                    itemptr := itemptr^.nextinqueue
                  END {while}
              END; {with entry}
            printentry(right,position)
          END {with subtree^}
    END; {printentry}

  BEGIN {printtree}
    numberwidth := trucnc(ln(maxlines)/ln(radix)) + 1 + numbergap;
    maxonline := (maxlinelen - maxwordlen) DIV numberwidth;
    pageposition := maxonpage;
    printentry(tree,pageposition);
  END; {printtree}

BEGIN {crossreference}
  letters := ['a' .. 'z'];
  wordtree := NIL;
  buildtree(wordtree);
  printtree(wordtree);
END. {crossreference}

